@@ -22,10 +22,11 @@
 import urllib.request, json, sys, os, shutil, subprocess
 import configparser, hashlib
 
+req_timeout = 600.0
 private_repos = {'meson', 'wrapweb', 'meson-ci'}
 
 def gh_get(url):
-    r = urllib.request.urlopen(url)
+    r = urllib.request.urlopen(url, timeout=req_timeout)
     jd = json.loads(r.read().decode('utf-8'))
     return jd
 
@@ -39,13 +40,13 @@ def list_projects():
     return 0
 
 def unpack(sproj, branch, outdir):
-    subprocess.check_call(['git', 'clone', '-b', branch, 'https://github.com/mesonbuild/%s.git' % sproj, outdir])
+    subprocess.check_call(['git', 'clone', '-b', branch, 'https://github.com/mesonbuild/{}.git'.format(sproj), outdir])
     usfile = os.path.join(outdir, 'upstream.wrap')
     assert(os.path.isfile(usfile))
     config = configparser.ConfigParser()
     config.read(usfile)
     us_url = config['wrap-file']['source_url']
-    us = urllib.request.urlopen(us_url).read()
+    us = urllib.request.urlopen(us_url, timeout=req_timeout).read()
     h = hashlib.sha256()
     h.update(us)
     dig = h.hexdigest()
@@ -81,7 +82,7 @@ def install(sproj):
     if os.path.isdir(sproj_dir):
         print('Subproject is already there. To update, nuke the dir and reinstall.')
         return 1
-    blist = gh_get('https://api.github.com/repos/mesonbuild/%s/branches' % sproj)
+    blist = gh_get('https://api.github.com/repos/mesonbuild/{}/branches'.format(sproj))
     blist = [b['name'] for b in blist]
     blist = [b for b in blist if b != 'master']
     blist.sort()
