@@ -14,7 +14,7 @@
 
 # Public symbols for compilers sub-package when using 'from . import compilers'
 __all__ = [
-    'CompilerType',
+    'Compiler',
 
     'all_languages',
     'base_options',
@@ -22,7 +22,6 @@ __all__ = [
     'clink_langs',
     'c_suffixes',
     'cpp_suffixes',
-    'get_macos_dylib_install_name',
     'get_base_compile_args',
     'get_base_link_args',
     'is_assembly',
@@ -32,9 +31,10 @@ __all__ = [
     'is_object',
     'is_source',
     'lang_suffixes',
-    'sanitizer_compile_args',
     'sort_clink',
 
+    'AppleClangCCompiler',
+    'AppleClangCPPCompiler',
     'ArmCCompiler',
     'ArmCPPCompiler',
     'ArmclangCCompiler',
@@ -55,49 +55,58 @@ __all__ = [
     'G95FortranCompiler',
     'GnuCCompiler',
     'ElbrusCCompiler',
+    'EmscriptenCCompiler',
     'GnuCompiler',
     'GnuCPPCompiler',
     'ElbrusCPPCompiler',
+    'EmscriptenCPPCompiler',
     'GnuDCompiler',
     'GnuFortranCompiler',
     'ElbrusFortranCompiler',
+    'FlangFortranCompiler',
     'GnuObjCCompiler',
     'GnuObjCPPCompiler',
-    'IntelCompiler',
+    'IntelGnuLikeCompiler',
+    'IntelVisualStudioLikeCompiler',
     'IntelCCompiler',
     'IntelCPPCompiler',
+    'IntelClCCompiler',
+    'IntelClCPPCompiler',
     'IntelFortranCompiler',
+    'IntelClFortranCompiler',
     'JavaCompiler',
     'LLVMDCompiler',
     'MonoCompiler',
+    'CudaCompiler',
     'VisualStudioCsCompiler',
     'NAGFortranCompiler',
     'ObjCCompiler',
     'ObjCPPCompiler',
     'Open64FortranCompiler',
     'PathScaleFortranCompiler',
+    'PGICCompiler',
+    'PGICPPCompiler',
     'PGIFortranCompiler',
     'RustCompiler',
     'CcrxCCompiler',
-    'CcrxCompiler',
     'CcrxCPPCompiler',
     'SunFortranCompiler',
     'SwiftCompiler',
     'ValaCompiler',
+    'VisualStudioLikeCompiler',
     'VisualStudioCCompiler',
     'VisualStudioCPPCompiler',
 ]
 
 # Bring symbols from each module into compilers sub-package namespace
 from .compilers import (
-    CompilerType,
+    Compiler,
     all_languages,
     base_options,
     clib_langs,
     clink_langs,
     c_suffixes,
     cpp_suffixes,
-    get_macos_dylib_install_name,
     get_base_compile_args,
     get_base_link_args,
     is_header,
@@ -107,35 +116,38 @@ from .compilers import (
     is_object,
     is_library,
     lang_suffixes,
-    sanitizer_compile_args,
     sort_clink,
-    ClangCompiler,
     CompilerArgs,
-    GnuCompiler,
-    IntelCompiler,
-    CcrxCompiler,
 )
 from .c import (
     CCompiler,
+    AppleClangCCompiler,
     ArmCCompiler,
     ArmclangCCompiler,
     ClangCCompiler,
     ClangClCCompiler,
     GnuCCompiler,
     ElbrusCCompiler,
+    EmscriptenCCompiler,
     IntelCCompiler,
+    IntelClCCompiler,
+    PGICCompiler,
     CcrxCCompiler,
     VisualStudioCCompiler,
 )
 from .cpp import (
     CPPCompiler,
+    AppleClangCPPCompiler,
     ArmCPPCompiler,
     ArmclangCPPCompiler,
     ClangCPPCompiler,
     ClangClCPPCompiler,
     GnuCPPCompiler,
     ElbrusCPPCompiler,
+    EmscriptenCPPCompiler,
     IntelCPPCompiler,
+    IntelClCPPCompiler,
+    PGICPPCompiler,
     CcrxCPPCompiler,
     VisualStudioCPPCompiler,
 )
@@ -146,12 +158,15 @@ from .d import (
     GnuDCompiler,
     LLVMDCompiler,
 )
+from .cuda import CudaCompiler
 from .fortran import (
     FortranCompiler,
     G95FortranCompiler,
     GnuFortranCompiler,
     ElbrusFortranCompiler,
+    FlangFortranCompiler,
     IntelFortranCompiler,
+    IntelClFortranCompiler,
     NAGFortranCompiler,
     Open64FortranCompiler,
     PathScaleFortranCompiler,
@@ -172,3 +187,7 @@ from .objcpp import (
 from .rust import RustCompiler
 from .swift import SwiftCompiler
 from .vala import ValaCompiler
+from .mixins.visualstudio import VisualStudioLikeCompiler
+from .mixins.gnu import GnuCompiler
+from .mixins.intel import IntelGnuLikeCompiler, IntelVisualStudioLikeCompiler
+from .mixins.clang import ClangCompiler
