@@ -15,14 +15,14 @@
 """Abstractions for the LLVM/Clang compiler family."""
 
 import os
-import typing
+import typing as T
 
 from ... import mesonlib
 from ...linkers import AppleDynamicLinker
 from ..compilers import clike_optimization_args
 from .gnu import GnuLikeCompiler
 
-if typing.TYPE_CHECKING:
+if T.TYPE_CHECKING:
     from ...environment import Environment
     from ...dependencies import Dependency  # noqa: F401
 
@@ -30,7 +30,7 @@ clang_color_args = {
     'auto': ['-Xclang', '-fcolor-diagnostics'],
     'always': ['-Xclang', '-fcolor-diagnostics'],
     'never': ['-Xclang', '-fno-color-diagnostics'],
-}  # type: typing.Dict[str, typing.List[str]]
+}  # type: T.Dict[str, T.List[str]]
 
 
 class ClangCompiler(GnuLikeCompiler):
@@ -45,22 +45,22 @@ class ClangCompiler(GnuLikeCompiler):
         # All Clang backends can also do LLVM IR
         self.can_compile_suffixes.add('ll')
 
-    def get_colorout_args(self, colortype: str) -> typing.List[str]:
+    def get_colorout_args(self, colortype: str) -> T.List[str]:
         return clang_color_args[colortype][:]
 
-    def get_optimization_args(self, optimization_level: str) -> typing.List[str]:
+    def get_optimization_args(self, optimization_level: str) -> T.List[str]:
         return clike_optimization_args[optimization_level]
 
     def get_pch_suffix(self) -> str:
         return 'pch'
 
-    def get_pch_use_args(self, pch_dir: str, header: str) -> typing.List[str]:
+    def get_pch_use_args(self, pch_dir: str, header: str) -> T.List[str]:
         # Workaround for Clang bug http://llvm.org/bugs/show_bug.cgi?id=15136
         # This flag is internal to Clang (or at least not documented on the man page)
         # so it might change semantics at any time.
         return ['-include-pch', os.path.join(pch_dir, self.get_pch_name(header))]
 
-    def has_multi_arguments(self, args: typing.List[str], env: 'Environment') -> typing.List[str]:
+    def has_multi_arguments(self, args: T.List[str], env: 'Environment') -> T.List[str]:
         myargs = ['-Werror=unknown-warning-option', '-Werror=unused-command-line-argument']
         if mesonlib.version_compare(self.version, '>=3.6.0'):
             myargs.append('-Werror=ignored-optimization-argument')
@@ -69,8 +69,8 @@ class ClangCompiler(GnuLikeCompiler):
             env)
 
     def has_function(self, funcname: str, prefix: str, env: 'Environment', *,
-                     extra_args: typing.Optional[typing.List[str]] = None,
-                     dependencies: typing.Optional[typing.List['Dependency']] = None) -> bool:
+                     extra_args: T.Optional[T.List[str]] = None,
+                     dependencies: T.Optional[T.List['Dependency']] = None) -> bool:
         if extra_args is None:
             extra_args = []
         # Starting with XCode 8, we need to pass this to force linker
@@ -83,7 +83,7 @@ class ClangCompiler(GnuLikeCompiler):
         return super().has_function(funcname, prefix, env, extra_args=extra_args,
                                     dependencies=dependencies)
 
-    def openmp_flags(self) -> typing.List[str]:
+    def openmp_flags(self) -> T.List[str]:
         if mesonlib.version_compare(self.version, '>=3.8.0'):
             return ['-fopenmp']
         elif mesonlib.version_compare(self.version, '>=3.7.0'):
