@@ -65,7 +65,7 @@ class DirMaker:
     def __enter__(self):
         return self
 
-    def __exit__(self, type, value, traceback):
+    def __exit__(self, exception_type, value, traceback):
         self.dirs.reverse()
         for d in self.dirs:
             append_to_log(self.lf, d)
@@ -93,7 +93,7 @@ def set_chown(path, user=None, group=None, dir_fd=None, follow_symlinks=True):
                                                  dir_fd=dir_fd,
                                                  follow_symlinks=follow_symlinks)
         shutil.chown(path, user, group)
-    except:
+    except Exception:
         raise
     finally:
         os.chown = real_os_chown
@@ -101,12 +101,12 @@ def set_chown(path, user=None, group=None, dir_fd=None, follow_symlinks=True):
 def set_chmod(path, mode, dir_fd=None, follow_symlinks=True):
     try:
         os.chmod(path, mode, dir_fd=dir_fd, follow_symlinks=follow_symlinks)
-    except (NotImplementedError, OSError, SystemError) as e:
+    except (NotImplementedError, OSError, SystemError):
         if not os.path.islink(path):
             os.chmod(path, mode, dir_fd=dir_fd)
 
 def sanitize_permissions(path, umask):
-    if umask is None:
+    if umask == 'preserve':
         return
     new_perms = 0o777 if is_executable(path, follow_symlinks=False) else 0o666
     new_perms &= ~umask
@@ -157,7 +157,7 @@ def restore_selinux_contexts():
     '''
     try:
         subprocess.check_call(['selinuxenabled'])
-    except (FileNotFoundError, PermissionError, subprocess.CalledProcessError) as e:
+    except (FileNotFoundError, NotADirectoryError, PermissionError, subprocess.CalledProcessError):
         # If we don't have selinux or selinuxenabled returned 1, failure
         # is ignored quietly.
         return
@@ -211,6 +211,7 @@ def check_for_stampfile(fname):
 class Installer:
 
     def __init__(self, options, lf):
+        self.did_install_something = False
         self.options = options
         self.lf = lf
 
@@ -323,7 +324,6 @@ class Installer:
                 # FIXME: what about symlinks?
                 self.do_copyfile(abs_src, abs_dst)
                 set_mode(abs_dst, install_mode, data.install_umask)
-                append_to_log(self.lf, abs_dst)
 
     def do_install(self, datafilename):
         with open(datafilename, 'rb') as ifile:
@@ -331,9 +331,10 @@ class Installer:
         d.destdir = os.environ.get('DESTDIR', '')
         d.fullprefix = destdir_join(d.destdir, d.prefix)
 
-        if d.install_umask is not None:
+        if d.install_umask != 'preserve':
             os.umask(d.install_umask)
 
+        self.did_install_something = False
         try:
             d.dirmaker = DirMaker(self.lf)
             with d.dirmaker:
@@ -344,6 +345,8 @@ class Installer:
                 self.install_data(d)
                 restore_selinux_contexts()
                 self.run_install_script(d)
+                if not self.did_install_something:
+                    print('Nothing to install.')
         except PermissionError:
             if shutil.which('pkexec') is not None and 'PKEXEC_UID' not in os.environ:
                 print('Installation failed due to insufficient permissions.')
@@ -355,6 +358,7 @@ class Installer:
 
     def install_subdirs(self, d):
         for (src_dir, dst_dir, mode, exclude) in d.install_subdirs:
+            self.did_install_something = True
             full_dst_dir = get_destdir_path(d, dst_dir)
             print('Installing subdir %s to %s' % (src_dir, full_dst_dir))
             d.dirmaker.makedirs(full_dst_dir, exist_ok=True)
@@ -362,6 +366,7 @@ class Installer:
 
     def install_data(self, d):
         for i in d.data:
+            self.did_install_something = True
             fullfilename = i[0]
             outfilename = get_destdir_path(d, i[1])
             mode = i[2]
@@ -372,6 +377,7 @@ class Installer:
 
     def install_man(self, d):
         for m in d.man:
+            self.did_install_something = True
             full_source_filename = m[0]
             outfilename = get_destdir_path(d, m[1])
             outdir = os.path.dirname(outfilename)
@@ -382,6 +388,7 @@ class Installer:
 
     def install_headers(self, d):
         for t in d.headers:
+            self.did_install_something = True
             fullfilename = t[0]
             fname = os.path.basename(fullfilename)
             outdir = get_destdir_path(d, t[1])
@@ -403,6 +410,7 @@ class Installer:
         child_env.update(env)
 
         for i in d.install_scripts:
+            self.did_install_something = True  # Custom script must report itself if it does nothing.
             script = i['exe']
             args = i['args']
             name = ' '.join(script + args)
@@ -417,6 +425,7 @@ class Installer:
 
     def install_targets(self, d):
         for t in d.targets:
+            self.did_install_something = True
             if not os.path.exists(t.fname):
                 # For example, import libraries of shared modules are optional
                 if t.optional:
@@ -443,18 +452,20 @@ class Installer:
                     if fname.endswith('.jar'):
                         print('Not stripping jar target:', os.path.basename(fname))
                         continue
-                    print('Stripping target {!r}'.format(fname))
+                    print('Stripping target {!r} using {}.'.format(fname, d.strip_bin[0]))
                     ps, stdo, stde = Popen_safe(d.strip_bin + [outname])
                     if ps.returncode != 0:
                         print('Could not strip file.\n')
                         print('Stdout:\n%s\n' % stdo)
                         print('Stderr:\n%s\n' % stde)
                         sys.exit(1)
-                pdb_filename = os.path.splitext(fname)[0] + '.pdb'
-                if not should_strip and os.path.exists(pdb_filename):
-                    pdb_outname = os.path.splitext(outname)[0] + '.pdb'
-                    self.do_copyfile(pdb_filename, pdb_outname)
-                    set_mode(pdb_outname, install_mode, d.install_umask)
+                if fname.endswith('.js'):
+                    # Emscripten outputs js files and optionally a wasm file.
+                    # If one was generated, install it as well.
+                    wasm_source = os.path.splitext(fname)[0] + '.wasm'
+                    if os.path.exists(wasm_source):
+                        wasm_output = os.path.splitext(outname)[0] + '.wasm'
+                        self.do_copyfile(wasm_source, wasm_output)
             elif os.path.isdir(fname):
                 fname = os.path.join(d.build_dir, fname.rstrip('/'))
                 outname = os.path.join(outdir, os.path.basename(fname))
@@ -492,7 +503,6 @@ def run(opts):
     log_dir = os.path.join(private_dir, '../meson-logs')
     if not os.path.exists(os.path.join(opts.wd, datafilename)):
         sys.exit('Install data not found. Run this command in build directory root.')
-    log_dir = os.path.join(private_dir, '../meson-logs')
     if not opts.no_rebuild:
         if not rebuild_all(opts.wd):
             sys.exit(-1)
