@@ -21,12 +21,12 @@ from . import mlog
 # This is the regex for the supported escape sequences of a regular string
 # literal, like 'abc\x00'
 ESCAPE_SEQUENCE_SINGLE_RE = re.compile(r'''
-    ( \\U........      # 8-digit hex escapes
-    | \\u....          # 4-digit hex escapes
-    | \\x..            # 2-digit hex escapes
-    | \\[0-7]{1,3}     # Octal escapes
-    | \\N\{[^}]+\}     # Unicode characters by name
-    | \\[\\'abfnrtv]   # Single-character escapes
+    ( \\U[A-Fa-f0-9]{8}   # 8-digit hex escapes
+    | \\u[A-Fa-f0-9]{4}   # 4-digit hex escapes
+    | \\x[A-Fa-f0-9]{2}   # 2-digit hex escapes
+    | \\[0-7]{1,3}        # Octal escapes
+    | \\N\{[^}]+\}        # Unicode characters by name
+    | \\[\\'abfnrtv]      # Single-character escapes
     )''', re.UNICODE | re.VERBOSE)
 
 class MesonUnicodeDecodeError(MesonException):
@@ -37,7 +37,7 @@ class MesonUnicodeDecodeError(MesonException):
 def decode_match(match):
     try:
         return codecs.decode(match.group(0), 'unicode_escape')
-    except UnicodeDecodeError as err:
+    except UnicodeDecodeError:
         raise MesonUnicodeDecodeError(match.group(0))
 
 class ParseException(MesonException):
@@ -90,8 +90,9 @@ class Lexer:
     def __init__(self, code):
         self.code = code
         self.keywords = {'true', 'false', 'if', 'else', 'elif',
-                         'endif', 'and', 'or', 'not', 'foreach', 'endforeach'}
-        self.future_keywords = {'continue', 'break', 'in', 'return'}
+                         'endif', 'and', 'or', 'not', 'foreach', 'endforeach',
+                         'in', 'continue', 'break'}
+        self.future_keywords = {'return'}
         self.token_specification = [
             # Need to be sorted longest to shortest.
             ('ignore', re.compile(r'[ \t]')),
@@ -189,7 +190,11 @@ This will become a hard error in a future Meson release.""", self.getline(line_s
                             line_start = mo.end() - len(lines[-1])
                     elif tid == 'number':
                         value = int(match_text, base=0)
-                    elif tid == 'eol' or tid == 'eol_cont':
+                    elif tid == 'eol_cont':
+                        lineno += 1
+                        line_start = loc
+                        break
+                    elif tid == 'eol':
                         lineno += 1
                         line_start = loc
                         if par_count > 0 or bracket_count > 0 or curl_count > 0:
@@ -207,7 +212,15 @@ This will become a hard error in a future Meson release.""", self.getline(line_s
             if not matched:
                 raise ParseException('lexer', self.getline(line_start), lineno, col)
 
-class ElementaryNode:
+class BaseNode:
+    def accept(self, visitor):
+        fname = 'visit_{}'.format(type(self).__name__)
+        if hasattr(visitor, fname):
+            func = getattr(visitor, fname)
+            if callable(func):
+                func(self)
+
+class ElementaryNode(BaseNode):
     def __init__(self, token):
         self.lineno = token.lineno
         self.subdir = token.subdir
@@ -242,28 +255,38 @@ class StringNode(ElementaryNode):
     def __str__(self):
         return "String node: '%s' (%d, %d)." % (self.value, self.lineno, self.colno)
 
-class ArrayNode:
-    def __init__(self, args):
+class ContinueNode(ElementaryNode):
+    pass
+
+class BreakNode(ElementaryNode):
+    pass
+
+class ArrayNode(BaseNode):
+    def __init__(self, args, lineno, colno, end_lineno, end_colno):
         self.subdir = args.subdir
-        self.lineno = args.lineno
-        self.colno = args.colno
+        self.lineno = lineno
+        self.colno = colno
+        self.end_lineno = end_lineno
+        self.end_colno = end_colno
         self.args = args
 
-class DictNode:
-    def __init__(self, args):
+class DictNode(BaseNode):
+    def __init__(self, args, lineno, colno, end_lineno, end_colno):
         self.subdir = args.subdir
-        self.lineno = args.lineno
-        self.colno = args.colno
+        self.lineno = lineno
+        self.colno = colno
+        self.end_lineno = end_lineno
+        self.end_colno = end_colno
         self.args = args
 
-class EmptyNode:
+class EmptyNode(BaseNode):
     def __init__(self, lineno, colno):
         self.subdir = ''
         self.lineno = lineno
         self.colno = colno
         self.value = None
 
-class OrNode:
+class OrNode(BaseNode):
     def __init__(self, left, right):
         self.subdir = left.subdir
         self.lineno = left.lineno
@@ -271,7 +294,7 @@ class OrNode:
         self.left = left
         self.right = right
 
-class AndNode:
+class AndNode(BaseNode):
     def __init__(self, left, right):
         self.subdir = left.subdir
         self.lineno = left.lineno
@@ -279,7 +302,7 @@ class AndNode:
         self.left = left
         self.right = right
 
-class ComparisonNode:
+class ComparisonNode(BaseNode):
     def __init__(self, ctype, left, right):
         self.lineno = left.lineno
         self.colno = left.colno
@@ -288,7 +311,7 @@ class ComparisonNode:
         self.right = right
         self.ctype = ctype
 
-class ArithmeticNode:
+class ArithmeticNode(BaseNode):
     def __init__(self, operation, left, right):
         self.subdir = left.subdir
         self.lineno = left.lineno
@@ -297,21 +320,21 @@ class ArithmeticNode:
         self.right = right
         self.operation = operation
 
-class NotNode:
+class NotNode(BaseNode):
     def __init__(self, location_node, value):
         self.subdir = location_node.subdir
         self.lineno = location_node.lineno
         self.colno = location_node.colno
         self.value = value
 
-class CodeBlockNode:
+class CodeBlockNode(BaseNode):
     def __init__(self, location_node):
         self.subdir = location_node.subdir
         self.lineno = location_node.lineno
         self.colno = location_node.colno
         self.lines = []
 
-class IndexNode:
+class IndexNode(BaseNode):
     def __init__(self, iobject, index):
         self.iobject = iobject
         self.index = index
@@ -319,7 +342,7 @@ class IndexNode:
         self.lineno = iobject.lineno
         self.colno = iobject.colno
 
-class MethodNode:
+class MethodNode(BaseNode):
     def __init__(self, subdir, lineno, colno, source_object, name, args):
         self.subdir = subdir
         self.lineno = lineno
@@ -329,32 +352,36 @@ class MethodNode:
         assert(isinstance(self.name, str))
         self.args = args
 
-class FunctionNode:
-    def __init__(self, subdir, lineno, colno, func_name, args):
+class FunctionNode(BaseNode):
+    def __init__(self, subdir, lineno, colno, end_lineno, end_colno, func_name, args):
         self.subdir = subdir
         self.lineno = lineno
         self.colno = colno
+        self.end_lineno = end_lineno
+        self.end_colno = end_colno
         self.func_name = func_name
         assert(isinstance(func_name, str))
         self.args = args
 
-class AssignmentNode:
-    def __init__(self, lineno, colno, var_name, value):
+class AssignmentNode(BaseNode):
+    def __init__(self, subdir, lineno, colno, var_name, value):
+        self.subdir = subdir
         self.lineno = lineno
         self.colno = colno
         self.var_name = var_name
         assert(isinstance(var_name, str))
         self.value = value
 
-class PlusAssignmentNode:
-    def __init__(self, lineno, colno, var_name, value):
+class PlusAssignmentNode(BaseNode):
+    def __init__(self, subdir, lineno, colno, var_name, value):
+        self.subdir = subdir
         self.lineno = lineno
         self.colno = colno
         self.var_name = var_name
         assert(isinstance(var_name, str))
         self.value = value
 
-class ForeachClauseNode:
+class ForeachClauseNode(BaseNode):
     def __init__(self, lineno, colno, varnames, items, block):
         self.lineno = lineno
         self.colno = colno
@@ -362,36 +389,37 @@ class ForeachClauseNode:
         self.items = items
         self.block = block
 
-class IfClauseNode:
+class IfClauseNode(BaseNode):
     def __init__(self, lineno, colno):
         self.lineno = lineno
         self.colno = colno
         self.ifs = []
         self.elseblock = EmptyNode(lineno, colno)
 
-class UMinusNode:
+class UMinusNode(BaseNode):
     def __init__(self, current_location, value):
         self.subdir = current_location.subdir
         self.lineno = current_location.lineno
         self.colno = current_location.colno
         self.value = value
 
-class IfNode:
+class IfNode(BaseNode):
     def __init__(self, lineno, colno, condition, block):
         self.lineno = lineno
         self.colno = colno
         self.condition = condition
         self.block = block
 
-class TernaryNode:
-    def __init__(self, lineno, colno, condition, trueblock, falseblock):
+class TernaryNode(BaseNode):
+    def __init__(self, subdir, lineno, colno, condition, trueblock, falseblock):
+        self.subdir = subdir
         self.lineno = lineno
         self.colno = colno
         self.condition = condition
         self.trueblock = trueblock
         self.falseblock = falseblock
 
-class ArgumentNode:
+class ArgumentNode(BaseNode):
     def __init__(self, token):
         self.lineno = token.lineno
         self.colno = token.colno
@@ -436,7 +464,9 @@ comparison_map = {'equal': '==',
                   'lt': '<',
                   'le': '<=',
                   'gt': '>',
-                  'ge': '>='
+                  'ge': '>=',
+                  'in': 'in',
+                  'notin': 'not in',
                   }
 
 # Recursive descent parser for Meson's definition language.
@@ -501,13 +531,13 @@ class Parser:
             value = self.e1()
             if not isinstance(left, IdNode):
                 raise ParseException('Plusassignment target must be an id.', self.getline(), left.lineno, left.colno)
-            return PlusAssignmentNode(left.lineno, left.colno, left.value, value)
+            return PlusAssignmentNode(left.subdir, left.lineno, left.colno, left.value, value)
         elif self.accept('assign'):
             value = self.e1()
             if not isinstance(left, IdNode):
                 raise ParseException('Assignment target must be an id.',
                                      self.getline(), left.lineno, left.colno)
-            return AssignmentNode(left.lineno, left.colno, left.value, value)
+            return AssignmentNode(left.subdir, left.lineno, left.colno, left.value, value)
         elif self.accept('questionmark'):
             if self.in_ternary:
                 raise ParseException('Nested ternary operators are not allowed.',
@@ -517,7 +547,7 @@ class Parser:
             self.expect('colon')
             falseblock = self.e1()
             self.in_ternary = False
-            return TernaryNode(left.lineno, left.colno, left, trueblock, falseblock)
+            return TernaryNode(left.subdir, left.lineno, left.colno, left, trueblock, falseblock)
         return left
 
     def e2(self):
@@ -543,6 +573,8 @@ class Parser:
         for nodename, operator_type in comparison_map.items():
             if self.accept(nodename):
                 return ComparisonNode(operator_type, left, self.e5())
+        if self.accept('not') and self.accept('in'):
+            return ComparisonNode('notin', left, self.e5())
         return left
 
     def e5(self):
@@ -594,7 +626,7 @@ class Parser:
             if not isinstance(left, IdNode):
                 raise ParseException('Function call must be applied to plain id',
                                      self.getline(), left.lineno, left.colno)
-            left = FunctionNode(left.subdir, left.lineno, left.colno, left.value, args)
+            left = FunctionNode(left.subdir, left.lineno, left.colno, self.current.lineno, self.current.colno, left.value, args)
         go_again = True
         while go_again:
             go_again = False
@@ -615,11 +647,11 @@ class Parser:
         elif self.accept('lbracket'):
             args = self.args()
             self.block_expect('rbracket', block_start)
-            return ArrayNode(args)
+            return ArrayNode(args, block_start.lineno, block_start.colno, self.current.lineno, self.current.colno)
         elif self.accept('lcurl'):
             key_values = self.key_values()
             self.block_expect('rcurl', block_start)
-            return DictNode(key_values)
+            return DictNode(key_values, block_start.lineno, block_start.colno, self.current.lineno, self.current.colno)
         else:
             return self.e9()
 
@@ -642,16 +674,8 @@ class Parser:
         a = ArgumentNode(s)
 
         while not isinstance(s, EmptyNode):
-            potential = self.current
             if self.accept('colon'):
-                if not isinstance(s, StringNode):
-                    raise ParseException('Key must be a string.',
-                                         self.getline(), s.lineno, s.colno)
-                if s.value in a.kwargs:
-                    # + 1 to colno to point to the actual string, not the opening quote
-                    raise ParseException('Duplicate dictionary key: {}'.format(s.value),
-                                         self.getline(), s.lineno, s.colno + 1)
-                a.set_kwarg(s.value, self.statement())
+                a.set_kwarg(s, self.statement())
                 potential = self.current
                 if not self.accept('comma'):
                     return a
@@ -754,6 +778,10 @@ class Parser:
             block = self.foreachblock()
             self.block_expect('endforeach', block_start)
             return block
+        if self.accept('continue'):
+            return ContinueNode(self.current)
+        if self.accept('break'):
+            return BreakNode(self.current)
         return self.statement()
 
     def codeblock(self):
