@@ -14,6 +14,7 @@
 
 # A tool to run tests in many different ways.
 
+from pathlib import Path
 from collections import namedtuple
 from copy import deepcopy
 import argparse
@@ -28,21 +29,20 @@ import pickle
 import platform
 import random
 import re
-import shlex
 import signal
 import subprocess
 import sys
 import tempfile
 import time
-import typing
+import typing as T
 
 from . import build
 from . import environment
 from . import mlog
 from .dependencies import ExternalProgram
-from .mesonlib import substring_is_in_list, MesonException
+from .mesonlib import MesonException, get_wine_shortpath, split_args
 
-if typing.TYPE_CHECKING:
+if T.TYPE_CHECKING:
     from .backend.backends import TestSerialisation
 
 # GNU autotools interprets a return code of 77 from tests it executes to
@@ -85,9 +85,11 @@ def add_arguments(parser: argparse.ArgumentParser) -> None:
                         help='Do not rebuild before running tests.')
     parser.add_argument('--gdb', default=False, dest='gdb', action='store_true',
                         help='Run test under gdb.')
+    parser.add_argument('--gdb-path', default='gdb', dest='gdb_path',
+                        help='Path to the gdb binary (default: gdb).')
     parser.add_argument('--list', default=False, dest='list', action='store_true',
                         help='List available tests.')
-    parser.add_argument('--wrapper', default=None, dest='wrapper', type=shlex.split,
+    parser.add_argument('--wrapper', default=None, dest='wrapper', type=split_args,
                         help='wrapper to run tests with (e.g. Valgrind)')
     parser.add_argument('-C', default='.', dest='wd',
                         help='directory to cd into before running')
@@ -115,7 +117,7 @@ def add_arguments(parser: argparse.ArgumentParser) -> None:
                         ' more time to execute.')
     parser.add_argument('--setup', default=None, dest='setup',
                         help='Which test setup to use.')
-    parser.add_argument('--test-args', default=[], type=shlex.split,
+    parser.add_argument('--test-args', default=[], type=split_args,
                         help='Arguments to pass to the specified test(s) or all tests')
     parser.add_argument('args', nargs='*',
                         help='Optional list of tests to run')
@@ -146,7 +148,7 @@ def returncode_to_status(retcode: int) -> str:
         signame = 'SIGinvalid'
     return '(exit status %d or signal %d %s)' % (retcode, signum, signame)
 
-def env_tuple_to_str(env: typing.Iterable[typing.Tuple[str, str]]) -> str:
+def env_tuple_to_str(env: T.Iterable[T.Tuple[str, str]]) -> str:
     return ''.join(["%s='%s' " % (k, v) for k, v in env])
 
 
@@ -185,11 +187,11 @@ class TAPParser:
     _RE_YAML_START = re.compile(r'(\s+)---.*')
     _RE_YAML_END = re.compile(r'\s+\.\.\.\s*')
 
-    def __init__(self, io: typing.Iterator[str]):
+    def __init__(self, io: T.Iterator[str]):
         self.io = io
 
-    def parse_test(self, ok: bool, num: int, name: str, directive: typing.Optional[str], explanation: typing.Optional[str]) -> \
-            typing.Generator[typing.Union['TAPParser.Test', 'TAPParser.Error'], None, None]:
+    def parse_test(self, ok: bool, num: int, name: str, directive: T.Optional[str], explanation: T.Optional[str]) -> \
+            T.Generator[T.Union['TAPParser.Test', 'TAPParser.Error'], None, None]:
         name = name.strip()
         explanation = explanation.strip() if explanation else None
         if directive is not None:
@@ -206,7 +208,7 @@ class TAPParser:
 
         yield self.Test(num, name, TestResult.OK if ok else TestResult.FAIL, explanation)
 
-    def parse(self) -> typing.Generator[typing.Union['TAPParser.Test', 'TAPParser.Error', 'TAPParser.Version', 'TAPParser.Plan', 'TAPParser.Bailout'], None, None]:
+    def parse(self) -> T.Generator[T.Union['TAPParser.Test', 'TAPParser.Error', 'TAPParser.Version', 'TAPParser.Plan', 'TAPParser.Bailout'], None, None]:
         found_late_test = False
         bailed_out = False
         plan = None
@@ -299,6 +301,9 @@ class TAPParser:
                     yield self.Version(version=version)
                 continue
 
+            if len(line) == 0:
+                continue
+
             yield self.Error('unexpected input at line %d' % (lineno,))
 
         if state == self._YAML:
@@ -314,10 +319,10 @@ class TAPParser:
 class TestRun:
 
     @classmethod
-    def make_exitcode(cls, test: 'TestSerialisation', test_env: typing.Dict[str, str],
-                      returncode: int, duration: float, stdo: typing.Optional[str],
-                      stde: typing.Optional[str],
-                      cmd: typing.Optional[typing.List[str]]) -> 'TestRun':
+    def make_exitcode(cls, test: 'TestSerialisation', test_env: T.Dict[str, str],
+                      returncode: int, starttime: float, duration: float,
+                      stdo: T.Optional[str], stde: T.Optional[str],
+                      cmd: T.Optional[T.List[str]]) -> 'TestRun':
         if returncode == GNU_SKIP_RETURNCODE:
             res = TestResult.SKIP
         elif returncode == GNU_ERROR_RETURNCODE:
@@ -326,12 +331,13 @@ class TestRun:
             res = TestResult.EXPECTEDFAIL if bool(returncode) else TestResult.UNEXPECTEDPASS
         else:
             res = TestResult.FAIL if bool(returncode) else TestResult.OK
-        return cls(test, test_env, res, returncode, duration, stdo, stde, cmd)
+        return cls(test, test_env, res, returncode, starttime, duration, stdo, stde, cmd)
 
     @classmethod
-    def make_tap(cls, test: 'TestSerialisation', test_env: typing.Dict[str, str],
-                 returncode: int, duration: float, stdo: str, stde: str,
-                 cmd: typing.Optional[typing.List[str]]) -> 'TestRun':
+    def make_tap(cls, test: 'TestSerialisation', test_env: T.Dict[str, str],
+                 returncode: int, starttime: float, duration: float,
+                 stdo: str, stde: str,
+                 cmd: T.Optional[T.List[str]]) -> 'TestRun':
         res = None
         num_tests = 0
         failed = False
@@ -364,15 +370,16 @@ class TestRun:
             else:
                 res = TestResult.FAIL if failed else TestResult.OK
 
-        return cls(test, test_env, res, returncode, duration, stdo, stde, cmd)
+        return cls(test, test_env, res, returncode, starttime, duration, stdo, stde, cmd)
 
-    def __init__(self, test: 'TestSerialisation', test_env: typing.Dict[str, str],
-                 res: TestResult, returncode: int, duration: float,
-                 stdo: typing.Optional[str], stde: typing.Optional[str],
-                 cmd: typing.Optional[typing.List[str]]):
+    def __init__(self, test: 'TestSerialisation', test_env: T.Dict[str, str],
+                 res: TestResult, returncode: int, starttime: float, duration: float,
+                 stdo: T.Optional[str], stde: T.Optional[str],
+                 cmd: T.Optional[T.List[str]]):
         assert isinstance(res, TestResult)
         self.res = res
         self.returncode = returncode
+        self.starttime = starttime
         self.duration = duration
         self.stdo = stdo
         self.stde = stde
@@ -386,7 +393,10 @@ class TestRun:
             res += 'NONE\n'
         else:
             test_only_env = set(self.env.items()) - set(os.environ.items())
-            res += '{}{}\n'.format(env_tuple_to_str(test_only_env), ' '.join(self.cmd))
+            starttime_str = time.strftime("%H:%M:%S", time.gmtime(self.starttime))
+            res += '{} {}{}\n'.format(
+                starttime_str, env_tuple_to_str(test_only_env), ' '.join(self.cmd)
+            )
         if self.stdo:
             res += '--- stdout ---\n'
             res += self.stdo
@@ -400,7 +410,7 @@ class TestRun:
         res += '-------\n\n'
         return res
 
-def decode(stream: typing.Union[None, bytes]) -> str:
+def decode(stream: T.Union[None, bytes]) -> str:
     if stream is None:
         return ''
     try:
@@ -408,14 +418,15 @@ def decode(stream: typing.Union[None, bytes]) -> str:
     except UnicodeDecodeError:
         return stream.decode('iso-8859-1', errors='ignore')
 
-def write_json_log(jsonlogfile: typing.TextIO, test_name: str, result: TestRun) -> None:
+def write_json_log(jsonlogfile: T.TextIO, test_name: str, result: TestRun) -> None:
     jresult = {'name': test_name,
                'stdout': result.stdo,
                'result': result.res.value,
+               'starttime': result.starttime,
                'duration': result.duration,
                'returncode': result.returncode,
                'env': result.env,
-               'command': result.cmd}  # type: typing.Dict[str, typing.Any]
+               'command': result.cmd}  # type: T.Dict[str, T.Any]
     if result.stde:
         jresult['stderr'] = result.stde
     jsonlogfile.write(json.dumps(jresult) + '\n')
@@ -425,39 +436,39 @@ def run_with_mono(fname: str) -> bool:
         return True
     return False
 
-def load_benchmarks(build_dir: str) -> typing.List['TestSerialisation']:
-    datafile = os.path.join(build_dir, 'meson-private', 'meson_benchmark_setup.dat')
-    if not os.path.isfile(datafile):
-        raise TestException('Directory ${!r} does not seem to be a Meson build directory.'.format(build_dir))
-    with open(datafile, 'rb') as f:
-        obj = typing.cast(typing.List['TestSerialisation'], pickle.load(f))
+def load_benchmarks(build_dir: str) -> T.List['TestSerialisation']:
+    datafile = Path(build_dir) / 'meson-private' / 'meson_benchmark_setup.dat'
+    if not datafile.is_file():
+        raise TestException('Directory {!r} does not seem to be a Meson build directory.'.format(build_dir))
+    with datafile.open('rb') as f:
+        obj = T.cast(T.List['TestSerialisation'], pickle.load(f))
     return obj
 
-def load_tests(build_dir: str) -> typing.List['TestSerialisation']:
-    datafile = os.path.join(build_dir, 'meson-private', 'meson_test_setup.dat')
-    if not os.path.isfile(datafile):
-        raise TestException('Directory ${!r} does not seem to be a Meson build directory.'.format(build_dir))
-    with open(datafile, 'rb') as f:
-        obj = typing.cast(typing.List['TestSerialisation'], pickle.load(f))
+def load_tests(build_dir: str) -> T.List['TestSerialisation']:
+    datafile = Path(build_dir) / 'meson-private' / 'meson_test_setup.dat'
+    if not datafile.is_file():
+        raise TestException('Directory {!r} does not seem to be a Meson build directory.'.format(build_dir))
+    with datafile.open('rb') as f:
+        obj = T.cast(T.List['TestSerialisation'], pickle.load(f))
     return obj
 
 
 class SingleTestRunner:
 
-    def __init__(self, test: 'TestSerialisation', test_env: typing.Dict[str, str],
-                 env: typing.Dict[str, str], options: argparse.Namespace):
+    def __init__(self, test: 'TestSerialisation', test_env: T.Dict[str, str],
+                 env: T.Dict[str, str], options: argparse.Namespace):
         self.test = test
         self.test_env = test_env
         self.env = env
         self.options = options
 
-    def _get_cmd(self) -> typing.Optional[typing.List[str]]:
+    def _get_cmd(self) -> T.Optional[T.List[str]]:
         if self.test.fname[0].endswith('.jar'):
             return ['java', '-jar'] + self.test.fname
         elif not self.test.is_cross_built and run_with_mono(self.test.fname[0]):
             return ['mono'] + self.test.fname
         else:
-            if self.test.is_cross_built:
+            if self.test.is_cross_built and self.test.needs_exe_wrapper:
                 if self.test.exe_runner is None:
                     # Can not run test on cross compiled executable
                     # because there is no execute wrapper.
@@ -475,27 +486,27 @@ class SingleTestRunner:
         cmd = self._get_cmd()
         if cmd is None:
             skip_stdout = 'Not run because can not execute cross compiled binaries.'
-            return TestRun(self.test, self.test_env, TestResult.SKIP, GNU_SKIP_RETURNCODE, 0.0, skip_stdout, None, None)
+            return TestRun(self.test, self.test_env, TestResult.SKIP, GNU_SKIP_RETURNCODE, time.time(), 0.0, skip_stdout, None, None)
         else:
             wrap = TestHarness.get_wrapper(self.options)
             if self.options.gdb:
                 self.test.timeout = None
             return self._run_cmd(wrap + cmd + self.test.cmd_args + self.options.test_args)
 
-    def _run_cmd(self, cmd: typing.List[str]) -> TestRun:
+    def _run_cmd(self, cmd: T.List[str]) -> TestRun:
         starttime = time.time()
 
         if len(self.test.extra_paths) > 0:
             self.env['PATH'] = os.pathsep.join(self.test.extra_paths + ['']) + self.env['PATH']
-            if substring_is_in_list('wine', cmd):
-                wine_paths = ['Z:' + p for p in self.test.extra_paths]
-                wine_path = ';'.join(wine_paths)
-                # Don't accidentally end with an `;` because that will add the
-                # current directory and might cause unexpected behaviour
-                if 'WINEPATH' in self.env:
-                    self.env['WINEPATH'] = wine_path + ';' + self.env['WINEPATH']
-                else:
-                    self.env['WINEPATH'] = wine_path
+            winecmd = []
+            for c in cmd:
+                winecmd.append(c)
+                if os.path.basename(c).startswith('wine'):
+                    self.env['WINEPATH'] = get_wine_shortpath(
+                        winecmd,
+                        ['Z:' + p for p in self.test.extra_paths] + self.env.get('WINEPATH', '').split(';')
+                    )
+                    break
 
         # If MALLOC_PERTURB_ is not set, or if it is set to an empty value,
         # (i.e., the test or the environment don't explicitly set it), set
@@ -528,7 +539,7 @@ class SingleTestRunner:
                 # We don't want setsid() in gdb because gdb needs the
                 # terminal in order to handle ^C and not show tcsetpgrp()
                 # errors avoid not being able to use the terminal.
-                os.setsid()
+                os.setsid()  # type: ignore
 
         p = subprocess.Popen(cmd,
                              stdout=stdout,
@@ -565,11 +576,11 @@ class SingleTestRunner:
             # killing a process and all its children so we need
             # to roll our own.
             if is_windows():
-                subprocess.call(['taskkill', '/F', '/T', '/PID', str(p.pid)])
+                subprocess.run(['taskkill', '/F', '/T', '/PID', str(p.pid)])
             else:
                 try:
                     # Kill the process group that setsid() created.
-                    os.killpg(p.pid, signal.SIGKILL)
+                    os.killpg(p.pid, signal.SIGKILL)  # type: ignore
                 except ProcessLookupError:
                     # Sometimes (e.g. with Wine) this happens.
                     # There's nothing we can do (maybe the process
@@ -605,20 +616,20 @@ class SingleTestRunner:
             stdo = ""
             stde = additional_error
         if timed_out:
-            return TestRun(self.test, self.test_env, TestResult.TIMEOUT, p.returncode, duration, stdo, stde, cmd)
+            return TestRun(self.test, self.test_env, TestResult.TIMEOUT, p.returncode, starttime, duration, stdo, stde, cmd)
         else:
             if self.test.protocol == 'exitcode':
-                return TestRun.make_exitcode(self.test, self.test_env, p.returncode, duration, stdo, stde, cmd)
+                return TestRun.make_exitcode(self.test, self.test_env, p.returncode, starttime, duration, stdo, stde, cmd)
             else:
                 if self.options.verbose:
                     print(stdo, end='')
-                return TestRun.make_tap(self.test, self.test_env, p.returncode, duration, stdo, stde, cmd)
+                return TestRun.make_tap(self.test, self.test_env, p.returncode, starttime, duration, stdo, stde, cmd)
 
 
 class TestHarness:
     def __init__(self, options: argparse.Namespace):
         self.options = options
-        self.collected_logs = []  # type: typing.List[str]
+        self.collected_logs = []  # type: T.List[str]
         self.fail_count = 0
         self.expectedfail_count = 0
         self.unexpectedpass_count = 0
@@ -627,9 +638,9 @@ class TestHarness:
         self.timeout_count = 0
         self.is_run = False
         self.tests = None
-        self.logfilename = None   # type: typing.Optional[str]
-        self.logfile = None       # type: typing.Optional[typing.TextIO]
-        self.jsonlogfile = None   # type: typing.Optional[typing.TextIO]
+        self.logfilename = None   # type: T.Optional[str]
+        self.logfile = None       # type: T.Optional[T.TextIO]
+        self.jsonlogfile = None   # type: T.Optional[T.TextIO]
         if self.options.benchmark:
             self.tests = load_benchmarks(options.wd)
         else:
@@ -641,12 +652,23 @@ class TestHarness:
         self.suites = list(ss)
 
     def __del__(self) -> None:
+        self.close_logfiles()
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_value, traceback) -> None:
+        self.close_logfiles()
+
+    def close_logfiles(self) -> None:
         if self.logfile:
             self.logfile.close()
+            self.logfile = None
         if self.jsonlogfile:
             self.jsonlogfile.close()
+            self.jsonlogfile = None
 
-    def merge_suite_options(self, options: argparse.Namespace, test: 'TestSerialisation') -> typing.Dict[str, str]:
+    def merge_suite_options(self, options: argparse.Namespace, test: 'TestSerialisation') -> T.Dict[str, str]:
         if ':' in options.setup:
             if options.setup not in self.build_data.test_setups:
                 sys.exit("Unknown test setup '%s'." % options.setup)
@@ -698,7 +720,7 @@ class TestHarness:
         else:
             sys.exit('Unknown test result encountered: {}'.format(result.res))
 
-    def print_stats(self, numlen: int, tests: typing.List['TestSerialisation'],
+    def print_stats(self, numlen: int, tests: T.List['TestSerialisation'],
                     name: str, result: TestRun, i: int) -> None:
         startpad = ' ' * (numlen - len('%d' % (i + 1)))
         num = '%s%d/%d' % (startpad, i + 1, len(tests))
@@ -781,16 +803,16 @@ Timeout:            %4d
         return self.total_failure_count()
 
     @staticmethod
-    def split_suite_string(suite: str) -> typing.Tuple[str, str]:
+    def split_suite_string(suite: str) -> T.Tuple[str, str]:
         if ':' in suite:
             # mypy can't figure out that str.split(n, 1) will return a list of
             # length 2, so we have to help it.
-            return typing.cast(typing.Tuple[str, str], tuple(suite.split(':', 1)))
+            return T.cast(T.Tuple[str, str], tuple(suite.split(':', 1)))
         else:
             return suite, ""
 
     @staticmethod
-    def test_in_suites(test: 'TestSerialisation', suites: typing.List[str]) -> bool:
+    def test_in_suites(test: 'TestSerialisation', suites: T.List[str]) -> bool:
         for suite in suites:
             (prj_match, st_match) = TestHarness.split_suite_string(suite)
             for prjst in test.suite:
@@ -802,7 +824,7 @@ Timeout:            %4d
                 #     project_name:suite_name
                 # so we need to select only the test belonging to project_name
 
-                # this if hanlde the first case (i.e., SUITE == suite_name)
+                # this if handle the first case (i.e., SUITE == suite_name)
 
                 # in this way we can run tests belonging to different
                 # (sub)projects which share the same suite_name
@@ -822,10 +844,11 @@ Timeout:            %4d
         return False
 
     def test_suitable(self, test: 'TestSerialisation') -> bool:
-        return (not self.options.include_suites or TestHarness.test_in_suites(test, self.options.include_suites)) \
-            and not TestHarness.test_in_suites(test, self.options.exclude_suites)
+        return ((not self.options.include_suites or
+                TestHarness.test_in_suites(test, self.options.include_suites)) and not
+                TestHarness.test_in_suites(test, self.options.exclude_suites))
 
-    def get_tests(self) -> typing.List['TestSerialisation']:
+    def get_tests(self) -> T.List['TestSerialisation']:
         if not self.tests:
             print('No tests defined.')
             return []
@@ -838,6 +861,7 @@ Timeout:            %4d
         else:
             tests = self.tests
 
+        # allow specifying test names like "meson test foo1 foo2", where test('foo1', ...)
         if self.options.args:
             tests = [t for t in tests if t.name in self.options.args]
 
@@ -873,10 +897,10 @@ Timeout:            %4d
         self.logfile.write('Inherited environment: {}\n\n'.format(inherit_env))
 
     @staticmethod
-    def get_wrapper(options: argparse.Namespace) -> typing.List[str]:
-        wrap = []  # type: typing.List[str]
+    def get_wrapper(options: argparse.Namespace) -> T.List[str]:
+        wrap = []  # type: T.List[str]
         if options.gdb:
-            wrap = ['gdb', '--quiet', '--nh']
+            wrap = [options.gdb_path, '--quiet', '--nh']
             if options.repeat > 1:
                 wrap += ['-ex', 'run', '-ex', 'quit']
             # Signal the end of arguments to gdb
@@ -895,9 +919,9 @@ Timeout:            %4d
         else:
             return test.name
 
-    def run_tests(self, tests: typing.List['TestSerialisation']) -> None:
+    def run_tests(self, tests: T.List['TestSerialisation']) -> None:
         executor = None
-        futures = []  # type: typing.List[typing.Tuple[conc.Future[TestRun], int, typing.List[TestSerialisation], str, int]]
+        futures = []  # type: T.List[T.Tuple[conc.Future[TestRun], int, T.List[TestSerialisation], str, int]]
         numlen = len('%d' % len(tests))
         self.open_log_files()
         startdir = os.getcwd()
@@ -936,7 +960,7 @@ Timeout:            %4d
         finally:
             os.chdir(startdir)
 
-    def drain_futures(self, futures: typing.List[typing.Tuple['conc.Future[TestRun]', int, typing.List['TestSerialisation'], str, int]]) -> None:
+    def drain_futures(self, futures: T.List[T.Tuple['conc.Future[TestRun]', int, T.List['TestSerialisation'], str, int]]) -> None:
         for x in futures:
             (result, numlen, tests, name, i) = x
             if self.options.repeat > 1 and self.fail_count:
@@ -964,7 +988,7 @@ def list_tests(th: TestHarness) -> bool:
     return not tests
 
 def rebuild_all(wd: str) -> bool:
-    if not os.path.isfile(os.path.join(wd, 'build.ninja')):
+    if not (Path(wd) / 'build.ninja').is_file():
         print('Only ninja backend is supported to rebuild tests before running them.')
         return True
 
@@ -973,11 +997,9 @@ def rebuild_all(wd: str) -> bool:
         print("Can't find ninja, can't rebuild test.")
         return False
 
-    p = subprocess.Popen([ninja, '-C', wd])
-    p.communicate()
-
-    if p.returncode != 0:
-        print('Could not rebuild')
+    ret = subprocess.run([ninja, '-C', wd]).returncode
+    if ret != 0:
+        print('Could not rebuild {}'.format(wd))
         return False
 
     return True
@@ -1010,24 +1032,27 @@ def run(options: argparse.Namespace) -> int:
 
     if not options.list and not options.no_rebuild:
         if not rebuild_all(options.wd):
+            # We return 125 here in case the build failed.
+            # The reason is that exit code 125 tells `git bisect run` that the current commit should be skipped.
+            # Thus users can directly use `meson test` to bisect without needing to handle the does-not-build case separately in a wrapper script.
+            return 125
+
+    with TestHarness(options) as th:
+        try:
+            if options.list:
+                return list_tests(th)
+            if not options.args:
+                return th.doit()
+            return th.run_special()
+        except TestException as e:
+            print('Meson test encountered an error:\n')
+            if os.environ.get('MESON_FORCE_BACKTRACE'):
+                raise e
+            else:
+                print(e)
             return 1
 
-    try:
-        th = TestHarness(options)
-        if options.list:
-            return list_tests(th)
-        if not options.args:
-            return th.doit()
-        return th.run_special()
-    except TestException as e:
-        print('Meson test encountered an error:\n')
-        if os.environ.get('MESON_FORCE_BACKTRACE'):
-            raise e
-        else:
-            print(e)
-        return 1
-
-def run_with_args(args: typing.List[str]) -> int:
+def run_with_args(args: T.List[str]) -> int:
     parser = argparse.ArgumentParser(prog='meson test')
     add_arguments(parser)
     options = parser.parse_args(args)
