@@ -1,5 +1,3 @@
-#!/usr/bin/env python3
-
 # Copyright 2016 The Meson development team
 
 # Licensed under the Apache License, Version 2.0 (the "License");
@@ -16,7 +14,7 @@
 
 def destdir_join(d1, d2):
     # c:\destdir + c:\prefix must produce c:\destdir\prefix
-    if len(d1) > 1 and d1[1] == ':' and \
-        len(d2) > 1 and d2[1] == ':':
+    if len(d1) > 1 and d1[1] == ':' \
+            and len(d2) > 1 and d2[1] == ':':
         return d1 + d2[2:]
     return d1 + d2
