@@ -123,7 +123,7 @@ class Elf(DataSizes):
             self.parse_header()
             self.parse_sections()
             self.parse_dynamic()
-        except:
+        except (struct.error, RuntimeError):
             self.bf.close()
             raise
 
@@ -180,7 +180,7 @@ class Elf(DataSizes):
     def parse_sections(self):
         self.bf.seek(self.e_shoff)
         self.sections = []
-        for i in range(self.e_shnum):
+        for _ in range(self.e_shnum):
             self.sections.append(SectionHeader(self.bf, self.ptrsize, self.is_le))
 
     def read_str(self):
@@ -413,9 +413,8 @@ def fix_darwin(fname, new_rpath, final_path, install_name_mappings):
             subprocess.check_call(['install_name_tool', fname] + args,
                                   stdout=subprocess.DEVNULL,
                                   stderr=subprocess.DEVNULL)
-    except Exception:
-        raise
-        sys.exit(0)
+    except Exception as err:
+        raise SystemExit(err)
 
 def fix_jar(fname):
     subprocess.check_call(['jar', 'xfv', fname, 'META-INF/MANIFEST.MF'])
@@ -432,6 +431,9 @@ def fix_rpath(fname, new_rpath, final_path, install_name_mappings, verbose=True)
     # Static libraries never have rpaths
     if fname.endswith('.a'):
         return
+    # DLLs and EXE never have rpaths
+    if fname.endswith('.dll') or fname.endswith('.exe'):
+        return
     try:
         if fname.endswith('.jar'):
             fix_jar(fname)
