@@ -44,9 +44,9 @@ def read_linguas(src_sub):
         return []
 
 def run_potgen(src_sub, pkgname, datadirs, args):
-    listfile = os.path.join(src_sub, 'POTFILES')
+    listfile = os.path.join(src_sub, 'POTFILES.in')
     if not os.path.exists(listfile):
-        listfile = os.path.join(src_sub, 'POTFILES.in')
+        listfile = os.path.join(src_sub, 'POTFILES')
         if not os.path.exists(listfile):
             print('Could not find file POTFILES in %s' % src_sub)
             return 1
@@ -81,9 +81,11 @@ def do_install(src_sub, bld_sub, dest, pkgname, langs):
         srcfile = os.path.join(bld_sub, l + '.gmo')
         outfile = os.path.join(dest, l, 'LC_MESSAGES',
                                pkgname + '.mo')
-        os.makedirs(os.path.split(outfile)[0], exist_ok=True)
-        shutil.copyfile(srcfile, outfile)
-        shutil.copystat(srcfile, outfile)
+        tempfile = outfile + '.tmp'
+        os.makedirs(os.path.dirname(outfile), exist_ok=True)
+        shutil.copyfile(srcfile, tempfile)
+        shutil.copystat(srcfile, tempfile)
+        os.replace(tempfile, outfile)
         print('Installing %s to %s' % (srcfile, outfile))
     return 0
 
