@@ -1,7 +1,7 @@
 #import <stdio.h>
 #import "vis.h"
 
-int EXPORT_PUBLIC libfoo(int arg) {
+int EXPORT_PUBLIC libfoo(void) {
   return 0;
 }
 
